from collections.abc import (
    AsyncIterator,
)
from typing import (
    Any,
)

from ...database import (
    DatabaseClient,
)
from .operations import (
    MockedDatabaseOperation,
)


class MockedDatabaseClient(DatabaseClient):
    """For testing purposes"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.kwargs = kwargs
        self._response = tuple()

    async def _reset(self, **kwargs) -> None:
        """For testing purposes"""
        self._response = tuple()

    async def _execute(self, operation: MockedDatabaseOperation) -> None:
        """For testing purposes"""
        self._response = operation.response

    async def _fetch_all(self, *args, **kwargs) -> AsyncIterator[Any]:
        """For testing purposes"""
        for value in self._response:
            yield value



from collections.abc import (
    Iterable,
)
from typing import (
    Any,
    Optional,
)

from ...database import (
    DatabaseOperation,
)


class MockedDatabaseOperation(DatabaseOperation):
    """For testing purposes"""

    def __init__(self, content: str, response: Optional[Iterable[Any]] = tuple(), *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.content = content
        self.response = tuple(response)



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from collections.abc import (
    Callable,
    Hashable,
)

from cached_property import (
    cached_property,
)

from .pools import (
    Pool,
)


class Lock(ABC):
    """Lock base class."""

    key: Hashable

    def __init__(self, key: Hashable, *args, **kwargs):
        if not isinstance(key, Hashable):
            raise ValueError(f"The key must be hashable. Obtained: {key!r} ({type(key)})")

        self.key = key

    async def __aenter__(self) -> Lock:
        await self.acquire()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.release()

    @abstractmethod
    async def acquire(self) -> None:
        """Acquire the lock.

        :return: This method does not return anything.
        """

    @abstractmethod
    async def release(self):
        """Release the lock.

        :return: This method does not return anything.
        """

    @cached_property
    def hashed_key(self) -> int:
        """Get the hashed key.

        :return: An integer value.
        """
        if not isinstance(self.key, int):
            return hash(self.key)
        return self.key


class LockPool(Pool[Lock], ABC):
    """Lock Pool class."""

    acquire: Callable[[str], Lock]



from __future__ import (
    annotations,
)

import logging
from abc import (
    ABC,
    abstractmethod,
)
from asyncio import (
    wait_for,
)
from collections.abc import (
    AsyncIterator,
    Hashable,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
)

from ...builders import (
    BuildableMixin,
    Builder,
)
from ...config import (
    Config,
)
from ..operations import (
    ComposedDatabaseOperation,
    DatabaseOperation,
    DatabaseOperationFactory,
)
from .exceptions import (
    ProgrammingException,
)

if TYPE_CHECKING:
    from ..locks import (
        DatabaseLock,
    )

logger = logging.getLogger(__name__)


class DatabaseClient(ABC, BuildableMixin):
    """Database Client base class."""

    _factories: dict[type[DatabaseOperationFactory], type[DatabaseOperationFactory]]
    _lock: Optional[DatabaseLock]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._lock = None

    @classmethod
    def _from_config(cls, config: Config, name: Optional[str] = None, **kwargs) -> DatabaseClient:
        return super()._from_config(config, **config.get_database_by_name(name), **kwargs)

    async def is_valid(self, **kwargs) -> bool:
        """Check if the instance is valid.

        :return: ``True`` if it is valid or ``False`` otherwise.
        """
        return await self._is_valid(**kwargs)

    async def _is_valid(self, **kwargs) -> bool:
        return True

    async def _destroy(self) -> None:
        await self.reset()
        await super()._destroy()

    async def reset(self, **kwargs) -> None:
        """Reset the current instance status.

        :param kwargs: Additional named parameters.
        :return: This method does not return anything.
        """
        await self._destroy_lock()
        return await self._reset(**kwargs)

    @abstractmethod
    async def _reset(self, **kwargs) -> None:
        raise NotImplementedError

    async def execute(self, operation: DatabaseOperation) -> None:
        """Execute an operation.

        :param operation: The operation to be executed.
        :return: This method does not return anything.
        """
        if not isinstance(operation, DatabaseOperation):
            raise ValueError(f"The operation must be a {DatabaseOperation!r} instance. Obtained: {operation!r}")

        if operation.lock is not None:
            await self._create_lock(operation.lock)

        if isinstance(operation, ComposedDatabaseOperation):
            await wait_for(self._execute_composed(operation), operation.timeout)
        else:
            await wait_for(self._execute(operation), operation.timeout)

    async def _execute_composed(self, operation: ComposedDatabaseOperation) -> None:
        for op in operation.operations:
            await self.execute(op)

    @abstractmethod
    async def _execute(self, operation: DatabaseOperation) -> None:
        raise NotImplementedError

    async def _create_lock(self, lock: Hashable, *args, **kwargs):
        if self._lock is not None and self._lock.key == lock:
            return
        await self._destroy_lock()

        from ..locks import (
            DatabaseLock,
        )

        self._lock = DatabaseLock(self, lock, *args, **kwargs)
        await self._lock.acquire()

    async def _destroy_lock(self):
        if self._lock is not None:
            logger.debug(f"Destroying {self._lock!r}...")
            await self._lock.release()
            self._lock = None

    @property
    def lock(self) -> Optional[DatabaseLock]:
        """Get the lock.

        :return: A ``DatabaseLock`` instance.
        """
        return self._lock

    async def fetch_one(self) -> Any:
        """Fetch one value.

        :return: This method does not return anything.
        """
        try:
            return await self.fetch_all().__anext__()
        except StopAsyncIteration:
            raise ProgrammingException("There are not any value to be fetched.")

    def fetch_all(self) -> AsyncIterator[Any]:
        """Fetch all values with an asynchronous iterator.

        :return: This method does not return anything.
        """
        return self._fetch_all()

    @abstractmethod
    def _fetch_all(self, *args, **kwargs) -> AsyncIterator[Any]:
        raise NotImplementedError

    @classmethod
    def set_factory(cls, base: type[DatabaseOperationFactory], impl: type[DatabaseOperationFactory]) -> None:
        """Register an operation factory implementation for an operation factory interface.

        :param base: The operation factory interface.
        :param impl: The operation factory implementation.
        :return: This method does not return anything.
        """
        if not issubclass(base, DatabaseOperationFactory):
            raise ValueError(f"{base!r} must be a subclass of {DatabaseOperationFactory!r}")

        if not issubclass(impl, base):
            raise ValueError(f"{impl!r} must be a subclass of {base!r}")

        if not hasattr(cls, "_factories"):
            cls._factories = dict()

        cls._factories[base] = impl

    @classmethod
    def get_factory(cls, base: type[DatabaseOperationFactory]) -> DatabaseOperationFactory:
        """Get an operation factory implementation for an operation factory interface.

        :param base: The operation factory interface.
        :return: The operation factory implementation.
        """
        if not hasattr(cls, "_factories") or base not in cls._factories:
            raise ValueError(f"{cls!r} does not contain any registered factory implementation for {base!r}")

        return cls._factories[base]()


class DatabaseClientBuilder(Builder[DatabaseClient]):
    """Database Client Builder class."""

    def with_name(self, name: str) -> DatabaseClientBuilder:
        """Set name.

        :param name: The name to be added.
        :return: This method return the builder instance.
        """
        self.kwargs["name"] = name
        return self

    def with_config(self, config: Config) -> DatabaseClientBuilder:
        """Set config.

        :param config: The config to be set.
        :return: This method return the builder instance.
        """
        database_config = config.get_database_by_name(self.kwargs.get("name"))
        self.kwargs |= database_config
        return self


DatabaseClient.set_builder(DatabaseClientBuilder)



from __future__ import (
    annotations,
)

from collections.abc import (
    Hashable,
)
from typing import (
    Optional,
)

from ...locks import (
    Lock,
)
from ..clients import (
    DatabaseClient,
)
from .factories import (
    LockDatabaseOperationFactory,
)


class DatabaseLock(Lock):
    """Database Lock class."""

    def __init__(
        self,
        client: DatabaseClient,
        key: Hashable,
        *args,
        operation_factory: Optional[LockDatabaseOperationFactory] = None,
        **kwargs,
    ):
        super().__init__(key, *args, **kwargs)
        if operation_factory is None:
            operation_factory = client.get_factory(LockDatabaseOperationFactory)

        self.client = client
        self.operation_factory = operation_factory

    async def acquire(self) -> None:
        """Acquire the lock.

        :return: This method does not return anything.
        """
        operation = self.operation_factory.build_acquire(self.hashed_key)
        await self.client.execute(operation)

    async def release(self) -> None:
        """Release the lock.

        :return: This method does not return anything.
        """
        operation = self.operation_factory.build_release(self.hashed_key)
        await self.client.execute(operation)



from __future__ import (
    annotations,
)

import unittest
from abc import (
    ABC,
)
from contextlib import (
    suppress,
)
from pathlib import (
    Path,
)
from typing import (
    Any,
    Union,
)
from uuid import (
    uuid4,
)

from ..config import (
    Config,
)
from ..database import (
    DatabaseClient,
    DatabaseClientPool,
    ManagementDatabaseOperationFactory,
)
from ..injections import (
    DependencyInjector,
    InjectableMixin,
)
from ..pools import (
    PoolFactory,
)


class MinosTestCase(unittest.IsolatedAsyncioTestCase, ABC):
    CONFIG_FILE_PATH: Path

    def setUp(self) -> None:
        super().setUp()

        self.config = self.get_config()
        self.injector = DependencyInjector(self.config, self.get_injections())
        self.injector.wire_injections()

    def get_config(self) -> Config:
        return Config(self.get_config_file_path())

    def get_config_file_path(self) -> Path:
        return self.CONFIG_FILE_PATH

    def get_injections(self) -> list[Union[InjectableMixin, type[InjectableMixin], str]]:
        return []

    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        await self.injector.setup_injections()

    async def asyncTearDown(self) -> None:
        await self.injector.destroy_injections()
        await super().asyncTearDown()

    def tearDown(self) -> None:
        self.injector.unwire_injections()
        super().tearDown()

    def __getattr__(self, item: str) -> Any:
        if item != "injector":
            with suppress(Exception):
                return getattr(self.injector, item)
        raise AttributeError(f"{type(self).__name__!r} does not contain the {item!r} attribute.")


class DatabaseMinosTestCase(MinosTestCase, ABC):
    def setUp(self):
        self.base_config = Config(self.get_config_file_path())
        self._uuid = uuid4()
        self._test_db = {"database": f"test_db_{self._uuid.hex}"}
        super().setUp()

    def get_client(self) -> DatabaseClient:
        default_config = self.base_config.get_default_database()
        client = default_config.get("client")
        return client.from_config(self.base_config)

    def get_config(self) -> Config:
        config = Config(self.get_config_file_path())

        base_fn = config.get_databases

        def _fn():
            return {k: (v | self._test_db) for k, v in base_fn().items()}

        config.get_databases = _fn
        return config

    def get_injections(self) -> list[Union[InjectableMixin, type[InjectableMixin], str]]:
        return [PoolFactory.from_config(self.config, default_classes={"database": DatabaseClientPool})]

    async def asyncSetUp(self):
        await self._create_database(self._test_db)
        await super().asyncSetUp()

    async def asyncTearDown(self):
        await super().asyncTearDown()
        await self._drop_database(self._test_db)

    async def _create_database(self, test: dict[str, Any]) -> None:
        await self._drop_database(test)

        async with self.get_client() as client:
            operation = client.get_factory(ManagementDatabaseOperationFactory).build_create(test["database"])
            await client.execute(operation)

    async def _drop_database(self, test: dict[str, Any]) -> None:
        async with self.get_client() as client:
            operation = client.get_factory(ManagementDatabaseOperationFactory).build_delete(test["database"])
            await client.execute(operation)



from __future__ import (
    annotations,
)

import logging
import warnings
from abc import (
    ABC,
)
from asyncio import (
    gather,
    sleep,
)
from typing import (
    Any,
    AsyncContextManager,
    Generic,
    Optional,
    TypeVar,
)

from aiomisc import (
    PoolBase,
)
from aiomisc.pool import (
    ContextManager,
)

from .config import (
    Config,
)
from .exceptions import (
    MinosConfigException,
    MinosException,
)
from .injections import (
    Injectable,
)
from .setup import (
    SetupMixin,
)

logger = logging.getLogger(__name__)

P = TypeVar("P")


@Injectable("pool_factory")
class PoolFactory(SetupMixin):
    """Pool Factory class."""

    _pools: dict[tuple[str, ...], Pool]

    def __init__(self, config: Config, default_classes: dict[str, type[Pool]] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if default_classes is None:
            default_classes = dict()

        self._config = config
        self._default_classes = default_classes
        self._pools = dict()

    @classmethod
    def _from_config(cls, config: Config, **kwargs) -> PoolFactory:
        return cls(config, **kwargs)

    async def _destroy(self) -> None:
        await self._destroy_pools()
        await super()._destroy()

    async def _destroy_pools(self):
        logger.debug("Destroying pools...")
        futures = (pool.destroy() for pool in self._pools.values())
        await gather(*futures)
        logger.debug("Destroyed pools!")

    def get_pool(self, type_: str, identifier: Optional[str] = None, **kwargs) -> Pool:
        """Get a pool from the factory.

        :param type_: The type of the pool.
        :param identifier:  An optional key that identifies the pool.
        :param kwargs: Additional named arguments.
        :return: A ``Pool`` instance.
        """
        key = (type_, identifier)
        if key not in self._pools:
            logger.debug(f"Creating the {key!r} pool...")
            self._pools[key] = self._create_pool(type_, identifier=identifier, **kwargs)
        return self._pools[key]

    def _create_pool(self, type_: str, **kwargs) -> Pool:
        # noinspection PyTypeChecker
        pool_cls = self._get_pool_cls(type_)
        try:
            pool = pool_cls.from_config(self._config, **kwargs)
        except MinosConfigException:
            raise PoolException("The pool could not be built.")
        return pool

    def _get_pool_cls(self, type_: str) -> type[Pool]:
        pool_cls = self._default_classes.get(type_)

        if pool_cls is None:
            pool_cls = self._config.get_pools().get("types", dict()).get(type_)

        if pool_cls is None:
            raise PoolException(
                f"There is not any provided {type!r} to build pools that matches the given type: {type_!r}"
            )

        return pool_cls


class _PoolBase(PoolBase, ABC):
    def __init__(self, *args, maxsize: int = 10, recycle: Optional[int] = None, **kwargs):
        super().__init__(maxsize=maxsize, recycle=recycle)


class Pool(SetupMixin, _PoolBase, Generic[P], ABC):
    """Base class for Pool implementations in minos"""

    def __init__(self, *args, already_setup: bool = True, **kwargs):
        super().__init__(*args, already_setup=already_setup, **kwargs)

    # noinspection PyUnresolvedReferences
    async def __acquire(self) -> Any:  # pragma: no cover
        if self._instances.empty() and not self._semaphore.locked():
            await self._PoolBase__create_new_instance()

        instance = await self._instances.get()

        # noinspection PyBroadException
        try:
            result = await self._check_instance(instance)
        except Exception:
            logger.warning("Check instance %r failed", instance)
            self._PoolBase__recycle_instance(instance)
        else:
            if not result:
                self._PoolBase__recycle_instance(instance)
                return await self._PoolBase__acquire()

        self._used.add(instance)
        logger.debug(f"Acquired instance: {instance!r}")
        return instance

    # noinspection PyUnresolvedReferences
    async def __release(self, instance: Any) -> Any:  # pragma: no cover
        await self._release_instance(instance)
        await self._PoolBase__release(instance)
        logger.debug(f"Released instance: {instance!r}")

    def acquire(self, *args, **kwargs) -> AsyncContextManager[P]:
        """Acquire a new instance wrapped on an asynchronous context manager.

        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: An asynchronous context manager.
        """
        # noinspection PyUnresolvedReferences
        return ContextManager(self.__acquire, self.__release)

    async def _destroy(self) -> None:
        if len(self._used):
            logger.info("Waiting for instances releasing...")
            while len(self._used):
                await sleep(0.1)
            logger.info("Released instances!")

        await self.close()

    async def _check_instance(self, instance: P) -> bool:
        return True

    async def _release_instance(self, instance: P) -> None:
        pass


class MinosPool(Pool, Generic[P], ABC):
    """MinosPool class."""

    def __init__(self, *args, **kwargs):
        warnings.warn(f"{MinosPool!r} has been deprecated. Use {Pool} instead.", DeprecationWarning)
        super().__init__(*args, **kwargs)


class PoolException(MinosException):
    """Exception to be raised when some problem related with a pool happens."""



from collections.abc import (
    AsyncIterator,
)
from contextlib import (
    suppress,
)
from typing import (
    Any,
    Generic,
    Optional,
    TypeVar,
    get_args,
    get_origin,
)

from ..exceptions import (
    NotProvidedException,
)
from ..injections import (
    Inject,
)
from ..pools import (
    PoolException,
    PoolFactory,
)
from ..setup import (
    SetupMixin,
)
from .operations import (
    DatabaseOperation,
    DatabaseOperationFactory,
)
from .pools import (
    DatabaseClient,
    DatabaseClientPool,
)

GenericDatabaseOperationFactory = TypeVar("GenericDatabaseOperationFactory", bound=DatabaseOperationFactory)


class DatabaseMixin(SetupMixin, Generic[GenericDatabaseOperationFactory]):
    """Database Mixin class."""

    @Inject()
    def __init__(
        self,
        database_pool: Optional[DatabaseClientPool] = None,
        pool_factory: Optional[PoolFactory] = None,
        database_key: Optional[tuple[str]] = None,
        operation_factory: Optional[GenericDatabaseOperationFactory] = None,
        operation_factory_cls: Optional[type[GenericDatabaseOperationFactory]] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs, pool_factory=pool_factory)

        if database_pool is None and pool_factory is not None:
            database_pool = self._get_pool_from_factory(pool_factory, database_key)

        if not isinstance(database_pool, DatabaseClientPool):
            raise NotProvidedException(f"A {DatabaseClientPool!r} instance is required. Obtained: {database_pool}")

        self._pool = database_pool

        if operation_factory is None:
            if operation_factory_cls is None:
                operation_factory_cls = self._get_generic_operation_factory()
            if operation_factory_cls is not None:
                operation_factory = self.database_client_cls.get_factory(operation_factory_cls)

        self._operation_factory = operation_factory

    @staticmethod
    def _get_pool_from_factory(pool_factory: PoolFactory, database_key: Optional[tuple[str]]):
        if database_key is None:
            database_key = tuple()

        for identifier in database_key:
            with suppress(PoolException):
                return pool_factory.get_pool(type_="database", identifier=identifier)

        return pool_factory.get_pool(type_="database")

    @property
    def database_operation_factory(self) -> Optional[GenericDatabaseOperationFactory]:
        """Get the operation factory if any.

        :return: A ``OperationFactory`` if it has been set or ``None`` otherwise.
        """
        return self._operation_factory

    def _get_generic_operation_factory(self) -> Optional[type[GenericDatabaseOperationFactory]]:
        operation_factory_cls = None
        # noinspection PyUnresolvedReferences
        for base in self.__orig_bases__:
            origin = get_origin(base)
            if origin is None or not issubclass(origin, DatabaseMixin):
                continue
            args = get_args(base)
            operation_factory_cls = args[0]
            if not isinstance(operation_factory_cls, type) or not issubclass(
                operation_factory_cls, DatabaseOperationFactory
            ):
                raise TypeError(f"{type(self)!r} must contain a {DatabaseOperationFactory!r} as generic value.")
        return operation_factory_cls

    async def execute_on_database_and_fetch_one(self, operation: DatabaseOperation) -> Any:
        """Submit an Operation and get the first response.

        :param operation: The operation to be executed.
        :return: This method does not return anything.
        """
        async with self.database_pool.acquire() as client:
            await client.execute(operation)
            return await client.fetch_one()

    # noinspection PyUnusedLocal
    async def execute_on_database_and_fetch_all(
        self, operation: DatabaseOperation, streaming_mode: Optional[bool] = None
    ) -> AsyncIterator[tuple]:
        """Submit an Operation and return an asynchronous iterator.

        :param operation: The operation to be executed.
        :param streaming_mode: If ``True`` return the values in streaming directly from the database (keep an open
            database connection), otherwise preloads the full set of values on memory and then retrieves them.
        :return: This method does not return anything.
        """
        if streaming_mode is None:
            streaming_mode = False

        async with self.database_pool.acquire() as client:
            await client.execute(operation)
            async_iterable = client.fetch_all()
            if streaming_mode:
                async for value in async_iterable:
                    yield value
                return

            iterable = [value async for value in async_iterable]

        for value in iterable:
            yield value

    # noinspection PyUnusedLocal
    async def execute_on_database(self, operation: DatabaseOperation) -> None:
        """Submit an Operation.

        :param operation: The operation to be executed.
        :return: This method does not return anything.
        """
        async with self.database_pool.acquire() as client:
            return await client.execute(operation)

    @property
    def database_client_cls(self) -> type[DatabaseClient]:
        """Get the client's class.

        :return: A ``type`` instance that is subclass of ``DatabaseClient``.
        """
        return self.database_pool.client_cls

    @property
    def database_pool(self) -> DatabaseClientPool:
        """Get the database pool.

        :return: A ``DatabaseClientPool`` object.
        """
        return self._pool
